#pragma once

#ifdef __cplusplus
extern "C"
{
#endif

#include "defines.h"

#include <stddef.h>

    short RectWidth(Rect *);
    short RectHeight(Rect *);
    void CenterWindow(WindowPtr);
    void GcDie(const char *);
    void GcWarn(const char *);
    void GcInfo(const char *);
    void str2cpy(char *dst, const char *src, size_t size);
    void GcStripTrailingWhitespace(char *str);
    OSStatus GcIsAppearancePresent(Boolean *haveAppearance);

#ifdef __cplusplus
}
#endif